from functools import partial

from django.core.exceptions import ValidationError
from django.db.models.fields import NOT_PROVIDED

from nap.utils import digattr


class field(property):
    '''A base class to compare against.'''
    def __new__(cls, *args, **kwargs):
        '''
        Allow specifying keyword arguments when used as a decorator.
        '''
        if not args:
            return partial(field, **kwargs)
        return super(field, cls).__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        self.required = kwargs.pop('required', False)
        self.default = kwargs.pop('default', NOT_PROVIDED)
        super(field, self).__init__(*args, **kwargs)

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        return self.fget(instance._obj)

    def __set__(self, instance, value):
        if self.fset is None:
            return
        self.fset(instance._obj, value)


class context_field(field):
    '''Special case of field that allows access to the Mapper itself'''
    def __init__(self, *args, **kwargs):
        super(context_field, self).__init__(*args, **kwargs)
        self.context = kwargs

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        return self.fget(self, instance._obj)

    def __set__(self, instance, value):
        if self.fset is None:
            return
        return self.fset(self, instance._obj, value)


class Field(field):
    '''
    class V(DataMapper):
        foo = Field('bar', default=1)
    '''
    def __init__(self, name, default=NOT_PROVIDED, filters=None,
                 required=True):
        self.name = name
        self.default = default
        self.filters = filters or []
        self.required = required

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        value = getattr(instance._obj, self.name, self.default)
        for filt in self.filters:
            try:
                value = filt.from_python(value)
            except (TypeError, ValueError):
                raise ValidationError('Invalid value')
        return value

    def __set__(self, instance, value):
        for filt in self.filters[::-1]:
            value = filt.to_python(value)
        setattr(instance._obj, self.name, value)


class DigField(Field):

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        return digattr(instance._obj, self.name, self.default)

    def __set__(self, instance, value):
        raise NotImplementedError


class MapperField(Field):
    def __init__(self, *args, **kwargs):
        self.mapper = kwargs.pop('mapper')
        super(MapperField, self).__init__(*args, **kwargs)

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        value = super(MapperField, self).__get__(instance, cls)
        mapper = self.mapper()
        return mapper << value

    def __set__(self, instance, value):
        mapper = self.mapper(instance)
        mapper._update(value, update=True)



import json

from django.test import Client, TestCase


class JsonClient(Client):

    def _massage(self, kwargs):
        if 'json' in kwargs:
            kwargs['data'] = json.dumps(kwargs.pop('json'))
            kwargs.setdefault('content_type', 'application/json')
        return kwargs

    def get(self, *args, **kwargs):
        kwargs = self._massage(kwargs)
        return super(JsonClient, self).get(*args, **kwargs)

    def post(self, *args, **kwargs):
        kwargs = self._massage(kwargs)
        return super(JsonClient, self).post(*args, **kwargs)

    def put(self, *args, **kwargs):
        kwargs = self._massage(kwargs)
        return super(JsonClient, self).put(*args, **kwargs)

    def delete(self, *args, **kwargs):
        kwargs = self._massage(kwargs)
        return super(JsonClient, self).delete(*args, **kwargs)


class ApiTestCase(TestCase):

    def assertResponseCode(self, url, status_code, method='get', **kwargs):
        result = getattr(self.client, method)(url, **kwargs)
        self.assertEqual(result.status_code, status_code)
        return result



from __future__ import unicode_literals

import re
from collections import OrderedDict

from django.core.exceptions import SuspiciousOperation
from django.http import Http404, HttpResponse, JsonResponse  # NOQA
from django.utils.encoding import iri_to_uri
from django.utils.six.moves import http_client
from django.utils.six.moves.urllib.parse import urlparse

'''Add some missing HttpResponse sub-classes'''


STATUS_CODES = list(http_client.responses.items()) + [
    (308, 'PERMANENT REDIRECT'),
    (427, 'BAD GEOLOCATION'),
]
STATUS_CODES = tuple(sorted(STATUS_CODES))

STATUS = OrderedDict(STATUS_CODES)
# Set constant-like properties for reverse lookup
for code, label in STATUS_CODES:
    setattr(STATUS, re.sub(r'\W', '_', label.upper()), code)


class BaseHttpResponse(HttpResponse, Exception):
    '''
    A sub-class of HttpResponse that is also an Exception, allowing us to
    raise/catch it.

    With thanks to schinkel's repose.
    '''

#
# Success Responses (2xx)
#


class HttpResponseSuccess(BaseHttpResponse):
    '''A base class for all 2xx responses, so we can issubclass test.'''


class OK(HttpResponseSuccess):
    status_code = STATUS.OK


class Created(HttpResponseSuccess):
    status_code = STATUS.CREATED


class Accepted(HttpResponseSuccess):
    status_code = STATUS.ACCEPTED


class NoContent(HttpResponseSuccess):
    status_code = STATUS.NO_CONTENT


class ResetContent(HttpResponseSuccess):
    status_code = STATUS.RESET_CONTENT


class PartialContent(HttpResponseSuccess):
    status_code = STATUS.PARTIAL_CONTENT

#
# Redirection Responses (3xx)
#


class HttpResponseRedirection(BaseHttpResponse):
    '''A base class for all 3xx responses.'''


class LocationHeaderMixin(object):
    '''Many 3xx responses require a Location header'''
    def __init__(self, location, *args, **kwargs):
        super(LocationHeaderMixin, self).__init__(*args, **kwargs)
        parsed = urlparse(location)
        if parsed.scheme and parsed.scheme not in self.allowed_schemes:
            raise SuspiciousOperation(
                "Unsafe redirect to URL with protocol '%s'" % parsed.scheme
            )
        self['Location'] = iri_to_uri(location)

    url = property(lambda self: self['Location'])


class MultipleChoices(HttpResponseRedirection):
    status_code = STATUS.MULTIPLE_CHOICES


class MovedPermanently(LocationHeaderMixin, HttpResponseRedirection):
    status_code = STATUS.MOVED_PERMANENTLY


class Found(LocationHeaderMixin, HttpResponseRedirection):
    status_code = STATUS.FOUND


class SeeOther(LocationHeaderMixin, HttpResponseRedirection):
    status_code = STATUS.SEE_OTHER


class NotModified(HttpResponseRedirection):
    status_code = STATUS.NOT_MODIFIED


class UseProxy(LocationHeaderMixin, HttpResponseRedirection):
    status_code = STATUS.USE_PROXY


class TemporaryRedirect(HttpResponseRedirection):
    status_code = STATUS.TEMPORARY_REDIRECT


class PermanentRedirect(HttpResponseRedirection):
    status_code = STATUS.PERMANENT_REDIRECT

#
# Common ancestor for 4xx and 5xx responses
#


class HttpResponseError(BaseHttpResponse):
    '''Common base class for all error responses'''

#
# Client Error Responses (4xx)
#


class HttpResponseClientError(HttpResponseError):
    '''A base class for all 4xx responses.'''


class BadRequest(HttpResponseClientError):
    status_code = STATUS.BAD_REQUEST


# XXX Auth-Realm ?
class Unauthorized(HttpResponseClientError):
    status_code = STATUS.UNAUTHORIZED


class PaymentRequired(HttpResponseClientError):
    status_code = STATUS.PAYMENT_REQUIRED


class Forbidden(HttpResponseClientError):
    status_code = STATUS.FORBIDDEN


class NotFound(HttpResponseClientError):
    status_code = STATUS.NOT_FOUND


class MethodNotAllowed(HttpResponseClientError):
    def __init__(self, permitted_methods, *args, **kwargs):
        super(MethodNotAllowed, self).__init__(*args, **kwargs)
        self['Allow'] = ', '.join(permitted_methods)

    status_code = STATUS.METHOD_NOT_ALLOWED


class NotAcceptable(HttpResponseClientError):
    status_code = STATUS.NOT_ACCEPTABLE


class ProxyAuthenticationRequired(HttpResponseClientError):
    status_code = STATUS.PROXY_AUTHENTICATION_REQUIRED


class RequestTimeout(HttpResponseClientError):
    status_code = STATUS.REQUEST_TIMEOUT


class Conflict(HttpResponseClientError):
    status_code = STATUS.CONFLICT


class Gone(HttpResponseClientError):
    status_code = STATUS.GONE


class LengthRequired(HttpResponseClientError):
    status_code = STATUS.LENGTH_REQUIRED


class PreconditionFailed(HttpResponseClientError):
    status_code = STATUS.PRECONDITION_FAILED


class RequestEntityTooLarge(HttpResponseClientError):
    status_code = STATUS.REQUEST_ENTITY_TOO_LARGE


class RequestURITooLong(HttpResponseClientError):
    status_code = STATUS.REQUEST_URI_TOO_LONG


class UnsupportedMediaType(HttpResponseClientError):
    status_code = STATUS.UNSUPPORTED_MEDIA_TYPE


class RequestedRangeNotSatisfiable(HttpResponseClientError):
    status_code = STATUS.REQUESTED_RANGE_NOT_SATISFIABLE


class ExpectationFailed(HttpResponseClientError):
    status_code = STATUS.EXPECTATION_FAILED


class BadGeolocation(HttpResponseClientError):
    status_code = STATUS.BAD_GEOLOCATION

#
# Server Error (5xx)
#


class HttpResponseServerError(HttpResponseError):
    '''A base class for 5xx responses.'''


class InternalServerError(HttpResponseServerError):
    status_code = STATUS.INTERNAL_SERVER_ERROR


class NotImplemented(HttpResponseServerError):
    status_code = STATUS.NOT_IMPLEMENTED


class BadGateway(HttpResponseServerError):
    status_code = STATUS.BAD_GATEWAY


class ServiceUnavailable(HttpResponseServerError):
    status_code = STATUS.SERVICE_UNAVAILABLE


class GatewayTimeout(HttpResponseServerError):
    status_code = STATUS.GATEWAY_TIMEOUT


class HttpVersiontNotSupported(HttpResponseServerError):
    status_code = STATUS.HTTP_VERSION_NOT_SUPPORTED




'''
A wrapper for django shortcuts that adapts responses to be from nap.http
'''

from django.shortcuts import _get_queryset

from nap import http


def get_object_or_404(klass, *args, **kwargs):
    """
    Uses get() to return an object, or raises a NotFound exception if the object
    does not exist.

    klass may be a Model, Manager, or QuerySet object. All other passed
    arguments and keyword arguments are used in the get() query.

    Note: Like with get(), an MultipleObjectsReturned will be raised if more
    han one object is found.
    """
    queryset = _get_queryset(klass)
    try:
        return queryset.get(*args, **kwargs)
    except queryset.model.DoesNotExist:
        raise http.NotFound()


def get_list_or_404(klass, *args, **kwargs):
    """
    Uses filter() to return a list of objects, or raise a NotFound exception if
    the list is empty.

    klass may be a Model, Manager, or QuerySet object. All other passed
    arguments and keyword arguments are used in the filter() query.
    """
    queryset = _get_queryset(klass)
    obj_list = list(queryset.filter(*args, **kwargs))
    if not obj_list:
        raise http.NotFound()
    return obj_list




from .fields import DigField, Field, MapperField, field, context_field  # NOQA
from .mappers import DataMapper  # NOQA
from .models import ModelDataMapper  # NOQA
from .utils import DictObject  # NOQA



from __future__ import unicode_literals

from django.core.exceptions import ValidationError
from django.views.generic import View
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.list import MultipleObjectMixin

from .. import http
from ..utils import JsonMixin, flatten_errors


class NapView(View):
    '''
    Base view for Nap CBV.

    Catches any http exceptions raised, and returns them.
    '''

    def dispatch(self, *args, **kwargs):
        '''
        If any code raises one of our HTTP responses, we should catch and
        return it.
        '''
        try:
            return super(NapView, self).dispatch(*args, **kwargs)
        except http.BaseHttpResponse as exc:
            return exc


class MapperMixin(JsonMixin):
    '''
    Base class for generating JSON responses using DataMappers.
    '''
    response_class = http.JsonResponse
    content_type = 'application/json'
    mapper_class = None

    # Defaults for safety
    object = None
    object_list = None
    mapper = None

    ok_status = http.STATUS.OK
    accepted_status = http.STATUS.ACCEPTED
    created_status = http.STATUS.CREATED
    deleted_status = http.STATUS.NO_CONTENT
    error_status = http.STATUS.BAD_REQUEST

    def get_mapper(self, obj=None):
        '''
        Get the mapper to use for this request.

        Default action is to use self.mapper_class
        '''
        return self.mapper_class(obj)

    def empty_response(self, **kwargs):
        '''
        Helper method to return an empty response.
        '''
        kwargs.setdefault('safe', False)
        return self.response_class('', **kwargs)

    def single_response(self, **kwargs):
        '''
        Helper method to return a single object.

        If `object` is not passed, it will try to use `self.object`.  If
        `self.object` is not set, it will call `self.get_object()`.

        If `mapper` is not passed, it will try to use `self.mapper`.  If
        `self.mapper` is not set, it will call `self.get_mapper()`.

        Returns a `self.response_class` instance, passed ``mapper << obj``,
        along with `**kwargs`.
        '''
        obj = kwargs.pop('object', self.object)
        if obj is None:
            obj = self.get_object()

        mapper = kwargs.pop('mapper', self.mapper)
        if mapper is None:
            mapper = self.get_mapper(obj)

        return self.response_class(mapper << obj, **kwargs)

    def multiple_response(self, **kwargs):
        '''
        Helper method to return an iterable of objects.

        If `object_list` is not passed, it will try to use `self.obect_list`.
        If `self.object_list` is not set, it will call `self.get_object_list()`.


        If `mapper` is not passed, it will try to use `self.mapper`.  If
        `self.mapper` is not set, it will call `self.get_mapper()`.
        
        Returns a `self.response_class` instance, passed a list of ``mapper <<
        obj`` applied to each object, along with `**kwargs`.
        '''
        kwargs.setdefault('safe', False)

        object_list = kwargs.pop('object_list', self.object_list)
        if object_list is None:
            object_list = self.get_queryset()

        mapper = kwargs.pop('mapper', self.mapper)
        if mapper is None:
            mapper = self.get_mapper()

        return self.response_class([
            mapper << obj
            for obj in object_list
        ], **kwargs)

    def accepted_response(self):
        '''
        Shortcut to return an ``empty_response`` using a ``status`` of ``self.accepted_status``.
        '''
        return self.empty_response(status=self.accepted_status)

    def created_response(self):
        '''
        Shortcut to return a ``single_response`` using a ``status`` of ``self.created_status``.
        '''
        return self.single_response(status=self.created_status)

    def deleted_response(self):
        '''
        Shortcut to return an ``empty_response`` using a ``status`` of ``self.deleted_status``.
        '''
        return self.empty_response(status=self.deleted_status)

    def error_response(self, errors):
        '''
        Helper method to return ``self.respone_class`` with a ``status`` of ``self.error_status``.

        Will flatten ``errors`` using ``flatten_error``.
        '''
        return self.response_class(
            flatten_errors(errors),
            status=self.error_status,
        )


# List views
class ListMixin(MapperMixin, MultipleObjectMixin):

    def ok_response(self, **kwargs):
        '''
        Shortcut to return a ``multiple_response`` with a ``status`` of ``self.ok_status``.
        '''
        kwargs.setdefault('status', self.ok_status)
        return self.multiple_response(**kwargs)


class ListGetMixin(object):

    def get(self, request, *args, **kwargs):
        return self.ok_response()


class ListPostMixin(object):

    def post(self, request, *args, **kwargs):
        self.mapper = self.get_mapper(self.model())
        self.data = self.get_request_data()

        try:
            self.object = self.mapper._apply(self.data)
        except ValidationError as e:
            return self.post_invalid(e.error_dict)

        return self.post_valid()

    def post_invalid(self, errors):
        return self.error_response(errors)

    def post_valid(self):
        self.object.save()

        return self.created_response()


class BaseListView(ListMixin, NapView):
    pass


# Object views
class ObjectMixin(MapperMixin, SingleObjectMixin):

    def ok_response(self, **kwargs):
        kwargs.setdefault('status', self.ok_status)
        return self.single_response(**kwargs)


class ObjectGetMixin(object):

    def get(self, request, *args, **kwargs):
        return self.ok_response()


class ObjectPutMixin(object):

    def put(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.mapper = self.get_mapper(self.object)

        self.data = self.get_request_data({})

        try:
            self.mapper._apply(self.data)
        except ValidationError as e:
            return self.put_invalid(e.error_dict)

        return self.put_valid()

    def put_valid(self):
        self.object.save()
        return self.ok_response()

    def put_invalid(self, errors):
        return self.error_response(errors)


class ObjectPatchMixin(object):

    def patch(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.mapper = self.get_mapper(self.object)

        self.data = self.get_request_data({})

        try:
            self.mapper._patch(self.data)
        except ValidationError as e:
            return self.patch_invalid(e.error_dict)

        return self.patch_valid()

    def patch_valid(self):
        self.object.save()
        return self.ok_response()

    def patch_invalid(self, errors):
        return self.error_response(errors)


class ObjectDeleteMixin(object):

    def delete(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.mapper = self.get_mapper(self.object)

        return self.delete_valid()

    def delete_valid(self):
        self.object.delete()

        return self.deleted_response()


class BaseObjectView(ObjectMixin, NapView):
    pass



import requests


class RPCProxy(object):
    def __init__(self, client, name):
        self.client = client
        self.name = name

    def __call__(self, **kwargs):
        try:
            resp = self.client.session.post(self.client.endpoint,
                                            data=json.dumps(kwargs),
                                            headers={'X-Rpc-Action': self.name},
                                           )
        except:
            raise
        return resp.json()


class RPCClient(object):
    def __init__(self, endpoint):
        self.endpoint = endpoint
        self.session = requests.Session()

    def __getattr__(self, key):
        return RPCProxy(self, name)



from __future__ import unicode_literals

import json
from cgi import parse_header, parse_multipart

from django.http import QueryDict
from django.utils import six

try:
    from django.core.handlers.wsgi import ISO_8859_1
except ImportError:  # pre-1.7
    ISO_8859_1 = str('iso-8859-1')


def digattr(obj, attr, default=None):
    '''Perform template-style dotted lookup'''
    steps = attr.split('.')
    for step in steps:
        try:    # dict lookup
            obj = obj[step]
        except (TypeError, AttributeError, KeyError):
            try:    # attribute lookup
                obj = getattr(obj, step)
            except (TypeError, AttributeError):
                try:    # list index lookup
                    obj = obj[int(step)]
                except (IndexError, ValueError, KeyError, TypeError):
                    return default
        if callable(obj) and not getattr(obj, 'do_not_call_in_templates', False):
            obj = obj()
    return obj


class JsonMixin(object):
    '''
    Common methods for handling JSON request data.
    '''
    CONTENT_TYPES = ['application/json', 'text/json']

    JSON_ENCODER = None
    JSON_DECODER = None

    def get_request_data(self, default=None):
        '''Retrieve data from request'''
        from django.conf import settings
        c_type, c_data = parse_header(self.request.META.get('CONTENT_TYPE', ''))
        encoding = self.request.encoding or settings.DEFAULT_CHARSET

        if c_type in self.CONTENT_TYPES:
            if not self.request.body:
                return default
            return self.loads(self.request.body.decode(encoding))

        if self.request.method in ('PUT', 'PATCH'):
            if c_type == 'application/x-www-form-urlencoded':
                return QueryDict(self.request.body, encoding=encoding)
            elif c_type == 'multipart/form-data':
                return parse_multipart(self.request.body, c_data)

        return self.request.POST

    def dumps(self, data, **kwargs):
        '''How to parse content that matches our content types list.'''
        kwargs.setdefault('cls', self.JSON_ENCODER)
        return json.dumps(data, **kwargs)

    def loads(self, data, **kwargs):
        '''Serialise data for responses.'''
        kwargs.setdefault('cls', self.JSON_DECODER)
        return json.loads(data, **kwargs)


def flatten_errors(errors):
    '''
    Utility function to turn an ErrorDict into a dict of lists of strings.
    '''
    return {
        field: [
            error if isinstance(error, six.string_types) else (
                error.message % error.params if error.params else error.message
            )
            for error in errors
        ]
        for field, errors in errors.items()
    }
